#!/usr/bin/env python

# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2017, 2018. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

import argparse  # for ArgumentParser
import subprocess  # for Popen
import os
from datetime import datetime
import math
import socket

def find_machine_config():
    # Get the sockets and cores counts from lscpu
    try:
        lscpu_process = subprocess.Popen(
            args="lscpu",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True)
        lscpu_results = lscpu_process.stdout.read().splitlines()
    except ValueError:
        parser.error(
            'lscpu failed to run. Please provide values for sockets and cores.'
        )
    except OSError:
        parser.error(
            'lscpu failed to run. Please provide values for sockets and cores.'
        )

    try:
        sockets = int(
            ''.join(filter(str.isdigit,
                   next((s for s in lscpu_results if "Socket(s):" in s),
                        None))))
    except TypeError:
        parser.error(
            'Unable to parse lscpu. Please provide values for sockets and cores.'
        )

    try:
        cores = int(
            ''.join(filter(str.isdigit,
                   next((s for s in lscpu_results
                         if "Core(s) per socket:" in s), None))))
    except TypeError:
        parser.error(
            'Unable to parse lscpu. Please provide values for sockets and cores.'
        )

    # Get the accelerators count from nvidia-smi
    try:
        env_mod = os.environ.copy()
        env_mod["PATH"] = "/usr/sbin:" + env_mod["PATH"]
        nvidia_smi_process = subprocess.Popen(
            args=["nvidia-smi", "--query-gpu=count", "--format=csv,noheader"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True)
        nvidia_smi_results, error = nvidia_smi_process.communicate()
        if nvidia_smi_process.returncode != 0:
           parser.error(
             'nvidia-smi failed to run. No accelerators found in the ',socket.gethostname(),' machine.'
           )
        accelerators = int(nvidia_smi_results.splitlines()[0])
    except ValueError:
        parser.error(
            'nvidia-smi failed to run. No accelerators found in the ',socket.gethostname(),' machine.'
        )
    except OSError:
        parser.error(
            'nvidia-smi failed to run. No accelerators found in the ',socket.gethostname(),' machine.'
        )

    # Invalid usage checks.
    if (cores < 1):
        parser.error('Cores need to be greater than 0.')
    if (sockets * cores < accelerators):
        parser.error(
            'Invalid argument combination. The number of cores should ' +
            ' be greater or equal to the number of accelerators in each machine.'
        )

    return accelerators, cores, sockets

#~ Handles the set up of CLI parameters, and errors that may occur from invalid parameters
def read_args():
    #~ set up of the arguments
    parser = argparse.ArgumentParser(
        description='Create rank file for specified hosts.')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        '-w',
        '--workdir',
        dest='workdir',
        default="",
        help='Workdir of snaprun tool')
    parser.add_argument(
        '-a',
        '--accelerators',
        dest='accelerators',
        default=-1,
        type=int,
        help='Number of accelerators per node. If no option given the number in '
        + 'the current machine is assumed. Error on values larger than 16.')
    group.add_argument(
        "-m",
        "--machine_report",
        action="store_true",
        default=False,
        dest="machine_report",
        help=argparse.SUPPRESS
    )
    parser.add_argument(
        'hosts_cli', metavar='hosts', help='Comma seperated list of hosts.')
    args = parser.parse_args()

    #~ Transforming hosts CSV into a list
    hosts = args.hosts_cli.split(",")

    if args.machine_report:
        accelerators, cores, sockets = find_machine_config()
    else:
        # Dummy values
        accelerators, cores, sockets = (None, None, None)

    return accelerators, cores, sockets, hosts, args.machine_report, args.workdir

def makeRankfile(hosts, workdir):
    print("#%s by snapml_rank_gen" % datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
    print("#host = %s" % ",".join(hosts))
    accelerators = 0
    config = {}
    for idx in range(len(hosts)):
        hostfile = workdir.strip()+"/"+hosts[idx]+"_config.out"
        with open(hostfile, 'r') as fp:
            config_vals = fp.readline().split(",")
            acc = int(config_vals[0])
            cores = int(config_vals[1])
            sockets = int(config_vals[2])
            config[hosts[idx]] = { "acc": acc, "cores": cores, "sockets" : sockets }
            accelerators += acc
        
    print("#accelerators = %s" % accelerators)
    print("")

    # Return if no accelerators are found
    if (accelerators == 0):
      return

    rank_index = 0
    for key, value in config.items():
        host = key
        acc = value['acc']
        # skip hosts with no accelerators
        if acc == 0:
            continue
        cores = value['cores']
        sockets = value['sockets']
        instances_per_socket = int(math.ceil(acc / (1.0 * sockets)))
        cores_per_instance = cores // instances_per_socket

        # Make sure we don't go beyond accelerators count.
        ranks_this_host = 0 # Allow up to acc local ranks for this host
        for socket_idx in range(0, sockets):
           for _idx in range(instances_per_socket):
               if ranks_this_host < acc:
                   ranks_this_host += 1
               else:
                   break
               cores_begin = _idx * cores_per_instance
               cores_end = (_idx + 1) * cores_per_instance - 1

               print("rank %d=%s\t   slot=%d:%d-%d" % (rank_index, host,
                                                       socket_idx, cores_begin, cores_end))
               rank_index += 1

        print("")

def main():
    accelerators, cores, sockets, hosts, machine_report, workdir = read_args()
    
    if machine_report:
        print("%d,%d,%d" % (accelerators, cores, sockets))
    else:
        makeRankfile(hosts, workdir)

if __name__ == "__main__":
    main()
