# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

from pai4sk.decomposition.truncated_svd import TruncatedSVDSklearn
import warnings
import numbers
import numpy as np
from .utils import _can_allow_cuml, _is_cuml_algorithm, _do_not_allow_cuml

class TruncatedSVD(TruncatedSVDSklearn):
    """Dimensionality reduction using truncated SVD (aka LSA).

    This transformer performs linear dimensionality reduction by means of
    truncated singular value decomposition (SVD). Contrary to PCA, this
    estimator does not center the data before computing the singular value
    decomposition. This means it can work with scipy.sparse matrices
    efficiently.

    In particular, truncated SVD works on term count/tf-idf matrices as
    returned by the vectorizers in pai4sk.feature_extraction.text. In that
    context, it is known as latent semantic analysis (LSA).

    This estimator supports two algorithms: a fast randomized SVD solver, and
    a "naive" algorithm that uses ARPACK as an eigensolver on (X * X.T) or
    (X.T * X), whichever is more efficient.

    If the input data is cudf dataframe and if possible, then the accelerated
    TruncatedSVD algorithm from cuML will be used. Otherwise,
    scikit-learn's TruncatedSVD algorithm will be used.

    cuML in pai4sk is currently supported only

    |  (a) with python 3.6 and
    |  (b) without MPI.
    |  If TruncatedSVD from cuML is run, then the return values from the APIs
    will be cudf dataframe and cudf Series objects instead of the return types
    of scikit-learn API.

    Read more in the :ref:`User Guide <LSA>`.

    Parameters
    ----------
    n_components : int, default = 2
        Desired dimensionality of output data.
        Must be strictly less than the number of features.
        The default value is useful for visualisation. For LSA, a value of
        100 is recommended.

    algorithm : string, "arpack", "randomized", "cuml", "auto", "full" or "jacobi". default = "auto".
        SVD solver to use. Either "arpack" for the ARPACK wrapper in SciPy
        (scipy.sparse.linalg.svds), or "randomized" for the randomized
        algorithm due to Halko (2009) if cuml can't be used.

        "auto" will become "full" if the arguments satisfy
        some validations for using cuml. "auto" will become "randomized"
        if cuml is not used. `algorithm` should be one of "auto", "cuml",
        "full" and "jacobi" to use cuml.

    n_iter : int, optional (default 5)
        Number of iterations for randomized SVD solver. Not used by ARPACK.
        The default is larger than the default in `randomized_svd` to handle
        sparse matrices that may have large slowly decaying spectrum.

    random_state : int, RandomState instance or None, optional, default = None
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    tol : float, optional
        Tolerance for ARPACK. 0 means machine precision. Ignored by randomized
        SVD solver.

    use_gpu : boolean, Default is True
        If True, cuML will use GPU 0. Applicable only for cuML.

    Attributes
    ----------
    components_ : array of shape (n_components, n_features) or cudf dataframe

    explained_variance_ : array of shape (n_components,) or cudf Series object
        The variance of the training samples transformed by a projection to
        each component.

    explained_variance_ratio_ : array of shape (n_components,) or cudf Series object
        Percentage of variance explained by each of the selected components.

    singular_values_ : array of shape (n_components,) or cudf Series object
        The singular values corresponding to each of the selected components.
        The singular values are equal to the 2-norms of the ``n_components``
        variables in the lower-dimensional space.

    Examples
    --------
    >>> from pai4sk.decomposition import TruncatedSVD
    >>> from pai4sk.random_projection import sparse_random_matrix
    >>> X = sparse_random_matrix(100, 100, density=0.01, random_state=42)
    >>> svd = TruncatedSVD(n_components=5, n_iter=7, random_state=42)
    >>> svd.fit(X)  # doctest: +NORMALIZE_WHITESPACE
    TruncatedSVD(algorithm='randomized', n_components=5, n_iter=7,
            random_state=42, tol=0.0)
    >>> print(svd.explained_variance_ratio_)  # doctest: +ELLIPSIS
    [0.0606... 0.0584... 0.0497... 0.0434... 0.0372...]
    >>> print(svd.explained_variance_ratio_.sum())  # doctest: +ELLIPSIS
    0.249...
    >>> print(svd.singular_values_)  # doctest: +ELLIPSIS
    [2.5841... 2.5245... 2.3201... 2.1753... 2.0443...]

    See also
    --------
    PCA

    References
    ----------
    Finding structure with randomness: Stochastic algorithms for constructing
    approximate matrix decompositions
    Halko, et al., 2009 (arXiv:909) https://arxiv.org/pdf/0909.4061.pdf

    Notes
    -----
    SVD suffers from a problem called "sign indeterminacy", which means the
    sign of the ``components_`` and the output from transform depend on the
    algorithm and random state. To work around this, fit instances of this
    class to data once, then keep the instance around to do transformations.

    """

    __slots__ = ('_TruncatedSVDSklearn')

    def __init__(self, n_components=2, algorithm="auto", n_iter=5,
                 random_state=None, tol=0., use_gpu=True):
        self.algorithm = algorithm
        self.n_components = n_components
        self.n_iter = n_iter
        self.random_state = random_state
        self.tol = tol
        self.use_gpu = use_gpu
        self.components_ = None
        self.singular_values_ = None
        self.explained_variance_ = None
        self.explained_variance_ratio_ = None
        self.noise_variance_ = None
        self._cumlTSVD = None
        self._TruncatedSVDSklearn = super(TruncatedSVD, self)
        # Perform basic validations for enabling cuML first
        self._cuml_basic_validations = _can_allow_cuml()

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._TruncatedSVDSklearn = super(TruncatedSVD, self)

    def fit(self, X, y=None):
        """Fit LSI model on training data X.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features) or cudf dataframe
            Training data.

        y : Ignored

        Returns
        -------
        self : object
            Returns the transformer object.
            If TruncatedSVD from cuML is run, then this fit method saves the computed
            values as cudf dataframes and cudf Series objects instead of the
            results' types seen from scikit-learn's fit method.
        """
        self._validate_fit_args()

        if self._check_set_cuml(X):
            print("Info: Invoking cuML truncatedSVD fit...")
            self._cumlTSVD.fit(X)
            self._save_cuml_fit_results()
        else:
            _do_not_allow_cuml(X, self.algorithm, ["cuml", "full", "jacobi"])
            if self.algorithm == "auto":
                self.algorithm = "randomized"
            print("Info: Invoking scikit-learn truncatedSVD fit...")
            self._TruncatedSVDSklearn.fit(X, y)

        return self

    def fit_transform(self, X, y=None):
        """Fit LSI model to X and perform dimensionality reduction on X.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features) or cudf dataframe
            Training data.
            If TruncatedSVD from cuML is run, then this method saves the computed values
            as cudf dataframes and cudf Series objects instead of the
            results' types seen from scikit-learn's API.

        y : Ignored

        Returns
        -------
        X_new : array of shape (n_samples, n_components) or cudf dataframe
            Reduced version of X. This will always be a dense array.
        """

        self._validate_fit_args()

        if self._check_set_cuml(X):
            print("Info: Invoking cuML truncatedSVD fit_transform...")
            X_new = self._cumlTSVD.fit_transform(X)
            self._save_cuml_fit_results()
            return X_new
        else:
            _do_not_allow_cuml(X, self.algorithm, ["cuml", "full", "jacobi"])
            if self.algorithm == "auto":
                self.algorithm = "randomized"
            print("Info: Invoking scikit-learn truncatedSVD fit_transform...")
            return self._TruncatedSVDSklearn.fit_transform(X)

    def transform(self, X):
        """Perform dimensionality reduction on X.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features) or cudf dataframe
            New data.

        Returns
        -------
        X_new : array of shape (n_samples, n_components) or cudf dataframe
            Reduced version of X. This will always be dense.
            If TruncatedSVD from cuML is run, then this method returns cudf
            dataframe instead of the results' types seen from scikit-learn's
            transform method.
        """
        if self._cumlTSVD:
            return self._cumlTSVD.transform(X)
        else:
            return self._TruncatedSVDSklearn.transform(X)

    def inverse_transform(self, X):
        """Transform X back to its original space.

        Returns an array or cudf dataframe X_original whose transform would be X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_components) or cudf dataframe
            New data.

        Returns
        -------
        X_original : array of shape (n_samples, n_features) or cudf dataframe
            Note that this is always dense.
            If TruncatedSVD from cuML is run, then this method returns cudf
            dataframe instead of the results' types seen from scikit-learn's
            transform method.
        """
        if self._cumlTSVD:
            return self._cumlTSVD.inverse_transform(X)
        else:
            return self._TruncatedSVDSklearn.inverse_transform(X)

    def _check_set_cuml(self, X):
        """
           Check if cuML can be used for this truncated SVD
        """

        # Perform basic validations for enabling cuML first
        if not self._cuml_basic_validations:
            return False

        # Perform further validations for enabling cuML
        import cudf
        if not isinstance(X, cudf.DataFrame):
            warnings.warn("Currently cuML in pai4sk can work only on " \
                          "cudf.DataFrame.", UserWarning)
            return False

        if not self.use_gpu:
            warnings.warn("Conflicting arguments are seen. cudf.DataFrame" \
                          " is given as input and use_gpu is set to False.",
                          UserWarning)
            return False

        if not _is_cuml_algorithm(self.algorithm, ['full', 'jacobi']):
            return False

        if not self._cumlTSVD:
            from pai4sk import sml_utils
            ndevices = sml_utils.get_gpu_count()
            if ndevices == 0:
                warnings.warn("Currently cuML in pai4sk can work only with GPUs." \
                              "No GPUs are available.", UserWarning)
                return False

            if self.algorithm == "auto" or self.algorithm == "cuml":
                self.algorithm = "full"
            from cuml import TruncatedSVD as cumltsvd
            self._cumlTSVD = cumltsvd(n_components=self.n_components,
                                      tol=self.tol, n_iter=self.n_iter,
                                      random_state=self.random_state,
                                      algorithm=self.algorithm)
            warnings.warn("Using TruncatedSVD from cuML.", UserWarning)

        return True

    def _validate_fit_args(self):
        """Validate the parameters used by fit method
        """
        if not isinstance(self.n_iter, numbers.Number) or self.n_iter < 0:
            raise ValueError('Number of iterations should be a positive number,'
                             ' got %d instead' % self.n_iter)

        if not isinstance(self.tol, numbers.Number) or self.tol < 0:
            raise ValueError("Tolerance for stopping criteria must be "
                             "positive number; got (tol=%r)" % self.tol)

    def _save_cuml_fit_results(self):
        """Save components and singular values"""

        self.components_ = self._cumlTSVD.components_
        self.singular_values_ = self._cumlTSVD.singular_values_
        self.explained_variance_ = self._cumlTSVD.explained_variance_
        self.explained_variance_ratio_ = self._cumlTSVD.explained_variance_ratio_
        self.noise_variance_ = self._cumlTSVD.noise_variance_

