# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

from pai4sk.linear_model.logistic import LogisticRegressionSklearn
from pai4sk.linear_model.logistic import _check_solver
from ...utils.validation import check_X_y
from ...utils.multiclass import check_classification_targets
from ...exceptions import NotFittedError
import time
import warnings
import numbers
import numpy as np
from .utils import _check_snapml_types, _check_snapml_type, _is_gpu_matrix_type, _get_data_size, _check_uparams
from _snap_ml.utils import _is_mpi_enabled, _is_data_snapml_partition_type
_mpi_enabled = _is_mpi_enabled()

class LogisticRegression(LogisticRegressionSklearn):
    """Logistic Regression (aka logit, MaxEnt) classifier.

    In the multiclass case, the training algorithm uses the one-vs-rest (OvR)
    scheme if the 'multi_class' option is set to 'ovr', and uses the cross-
    entropy loss if the 'multi_class' option is set to 'multinomial'.
    (Currently the 'multinomial' option is supported only by the 'lbfgs',
    'sag' and 'newton-cg' solvers.)

    This class implements regularized logistic regression using the
    'liblinear' library, 'newton-cg', 'sag' and 'lbfgs' solvers. It can handle
    both dense and sparse input. Use C-ordered arrays or CSR matrices
    containing 64-bit floats for optimal performance; any other input format
    will be converted (and copied).

    The 'newton-cg', 'sag', and 'lbfgs' solvers support only L2 regularization
    with primal formulation. The 'liblinear' solver supports both L1 and L2
    regularization, with a dual formulation only for the L2 penalty.

    Read more in the :ref:`User Guide <logistic_regression>`.

    For SnapML solver this supports both local and distributed(MPI) method of execution.

    Parameters
    ----------
    penalty : str, 'l1' or 'l2', default: 'l2'
        Used to specify the norm used in the penalization. The 'newton-cg',
        'sag' and 'lbfgs' solvers support only l2 penalties.

        .. versionadded:: 0.19
           l1 penalty with SAGA solver (allowing 'multinomial' + L1)

    dual : bool, default: False
        Dual or primal formulation. Dual formulation is only implemented for
        l2 penalty with liblinear solver. Prefer dual=False when
        n_samples > n_features.

    tol : float, default: 1e-4
        Tolerance for stopping criteria.

    C : float, default: 1.0
        Inverse of regularization strength; must be a positive float.
        Like in support vector machines, smaller values specify stronger
        regularization.

    fit_intercept : bool, default: True
        Specifies if a constant (a.k.a. bias or intercept) should be
        added to the decision function.

    intercept_scaling : float, default 1.
        Useful only when the solver 'liblinear' is used
        and self.fit_intercept is set to True. In this case, x becomes
        [x, self.intercept_scaling],
        i.e. a "synthetic" feature with constant value equal to
        intercept_scaling is appended to the instance vector.
        The intercept becomes ``intercept_scaling * synthetic_feature_weight``.

        Note! the synthetic feature weight is subject to l1/l2 regularization
        as all other features.
        To lessen the effect of regularization on synthetic feature weight
        (and therefore on the intercept) intercept_scaling has to be increased.

    class_weight : dict or 'balanced', default: None
        Weights associated with classes in the form ``{class_label: weight}``.
        If not given, all classes are supposed to have weight one.

        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input data
        as ``n_samples / (n_classes * np.bincount(y))``.

        Note that these weights will be multiplied with sample_weight (passed
        through the fit method) if sample_weight is specified.

        .. versionadded:: 0.17
           *class_weight='balanced'*

    random_state : int, RandomState instance or None, optional, default: None
        The seed of the pseudo random number generator to use when shuffling
        the data.  If int, random_state is the seed used by the random number
        generator; If RandomState instance, random_state is the random number
        generator; If None, the random number generator is the RandomState
        instance used by `np.random`. Used when ``solver`` == 'sag' or
        'liblinear'.

    solver : str, {'newton-cg', 'lbfgs', 'liblinear', 'sag', 'saga', 'snapml'}, \
             default: 'snapml', if 'snap_ml' library is in PYTHONPATH, else,
             default: 'liblinear'.

        Algorithm to use in the optimization problem.

        - For small datasets, 'liblinear' is a good choice, whereas 'sag' and
          'saga' are faster for large ones.
        - For multiclass problems, only 'newton-cg', 'sag', 'saga' and 'lbfgs'
          handle multinomial loss; 'liblinear' is limited to one-versus-rest
          schemes.
        - 'newton-cg', 'lbfgs' and 'sag' only handle L2 penalty, whereas
          'liblinear' and 'saga' handle L1 penalty.

        Note that 'sag' and 'saga' fast convergence is only guaranteed on
        features with approximately the same scale. You can
        preprocess the data with a scaler from pai4sk.preprocessing.

        .. versionadded:: 0.17
           Stochastic Average Gradient descent solver.
        .. versionadded:: 0.19
           SAGA solver.
        .. versionchanged:: 0.20
            Default will change from 'liblinear' to 'lbfgs' in 0.22.

    max_iter : int, default: 100
        Useful only for the newton-cg, sag and lbfgs solvers.
        Maximum number of iterations taken for the solvers to converge.

    multi_class : str, {'ovr', 'multinomial', 'auto'}, default: 'ovr'
        If the option chosen is 'ovr', then a binary problem is fit for each
        label. For 'multinomial' the loss minimised is the multinomial loss fit
        across the entire probability distribution, *even when the data is
        binary*. 'multinomial' is unavailable when solver='liblinear'.
        'auto' selects 'ovr' if the data is binary, or if solver='liblinear',
        and otherwise selects 'multinomial'.

        .. versionadded:: 0.18
           Stochastic Average Gradient descent solver for 'multinomial' case.
        .. versionchanged:: 0.20
            Default will change from 'ovr' to 'auto' in 0.22.

    verbose : int, default: 0
        For the liblinear and lbfgs solvers set verbose to any positive
        number for verbosity.

    warm_start : bool, default: False
        When set to True, reuse the solution of the previous call to fit as
        initialization, otherwise, just erase the previous solution.
        Useless for liblinear solver. See :term:`the Glossary <warm_start>`.

        .. versionadded:: 0.17
           *warm_start* to support *lbfgs*, *newton-cg*, *sag*, *saga* solvers.

    n_jobs : int or None, optional (default=None)
        Number of CPU cores used when parallelizing over classes if
        multi_class='ovr'". This parameter is ignored when the ``solver`` is
        set to 'liblinear' regardless of whether 'multi_class' is specified or
        not. ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
        context. ``-1`` means using all processors.
        See :term:`Glossary <n_jobs>` for more details.

    use_gpu : bool, default : True
        Flag for indicating the hardware platform used for training. If True, the training 
        is performed using the GPU. If False, the training is performed using the CPU.
        The value of this parameter is subjected to changed based on the training data unless set explicitly.
        Applicable only for snapml solver

    device_ids : array-like of int, default : []
        If use_gpu is True, it indicates the IDs of the GPUs used for training.
        For single-GPU training, set device_ids to the GPU ID to be used for training, e.g., [0].
        For multi-GPU training, set device_ids to a list of GPU IDs to be used for training, e.g., [0, 1].
        Applicable only for snapml solver

    num_threads : int, default : 1
        The number of threads used for running the training. The value of this parameter
        should be a multiple of 32 if the training is performed on GPU (use_gpu=True)
        (default value for GPU is 256). Applicable only for snapml solver

    return_training_history : str or None, default : None
        How much information about the training should be collected and returned by the fit function. By 
        default no information is returned (None), but this parameter can be set to "summary", to obtain 
        summary statistics at the end of training, or "full" to obtain a complete set of statistics 
        for the entire training procedure. Note, enabling either option will result in slower training.
        Applicable only for snapml solver

    privacy : bool, default : False
        Train the model using a differentially private algorithm.

    eta : float, default : 0.3
        Learning rate for the differentially private training algorithm.

    batch_size : int, default : 100
        Mini-batch size for the differentially private training algorithm.

    privacy_epsilon : float, default : 10.0
        Target privacy gaurantee. Learned model will be (privacy_epsilon, 0.01)-private.

    grad_clip : float, default: 1.0
        Gradient clipping parameter for the differentially private training algorithm

    Attributes
    ----------

    coef_ : array, shape (1, n_features) or (n_classes, n_features)
        Coefficient of the features in the decision function.

        `coef_` is of shape (1, n_features) when the given problem is binary.
        In particular, when `multi_class='multinomial'`, `coef_` corresponds
        to outcome 1 (True) and `-coef_` corresponds to outcome 0 (False).

    intercept_ : array, shape (1,) or (n_classes,)
        Intercept (a.k.a. bias) added to the decision function.

        If `fit_intercept` is set to False, the intercept is set to zero.
        `intercept_` is of shape (1,) when the given problem is binary.
        In particular, when `multi_class='multinomial'`, `intercept_`
        corresponds to outcome 1 (True) and `-intercept_` corresponds to
        outcome 0 (False).

    n_iter_ : array, shape (n_classes,) or (1, )
        Actual number of iterations for all classes. If binary or multinomial,
        it returns only 1 element. For liblinear solver, only the maximum
        number of iteration across all classes is given.

    training_history_ : dict
        It returns a dictionary with the following keys : 'epochs', 't_elap_sec', 'train_obj'.
        If 'return_training_history' is set to "summary", 'epochs' contains the total number of
        epochs performed, 't_elap_sec' contains the total time for completing all of those epochs.
        If 'return_training_history' is set to "full", 'epochs' indicates the number of epochs
        that have elapsed so far, and 't_elap_sec' contains the time to do those epochs.
        'train_obj' is the training loss.
        Applicable only for snapml solver.

    support_ : array-like
        Indices of the features that contribute to the decision. (only available for L1)

    model_sparsity_ : float
        Fraction of non-zeros in the model parameters. (only available for L1)

        .. versionchanged:: 0.20

            In SciPy <= 1.0.0 the number of lbfgs iterations may exceed
            ``max_iter``. ``n_iter_`` will now report at most ``max_iter``.

    Examples
    --------
    >>> from pai4sk.datasets import load_iris
    >>> from pai4sk.linear_model import LogisticRegression
    >>> X, y = load_iris(return_X_y=True)
    >>> clf = LogisticRegression(random_state=0, solver='lbfgs',
    ...                          multi_class='multinomial').fit(X, y)
    >>> clf.predict(X[:2, :])
    array([0, 0])
    >>> clf.predict_proba(X[:2, :]) # doctest: +ELLIPSIS
    array([[9.8...e-01, 1.8...e-02, 1.4...e-08],
           [9.7...e-01, 2.8...e-02, ...e-08]])
    >>> clf.score(X, y)
    0.97...

    See also
    --------
    SGDClassifier : incrementally trained logistic regression (when given
        the parameter ``loss="log"``).
    LogisticRegressionCV : Logistic regression with built-in cross validation

    Notes
    -----
    The underlying C implementation uses a random number generator to
    select features when fitting the model. It is thus not uncommon,
    to have slightly different results for the same input data. If
    that happens, try with a smaller tol parameter.

    Predict output may not match that of standalone liblinear in certain
    cases. See :ref:`differences from liblinear <liblinear_differences>`
    in the narrative documentation.

    References
    ----------

    LIBLINEAR -- A Library for Large Linear Classification
        http://www.csie.ntu.edu.tw/~cjlin/liblinear/

    SAG -- Mark Schmidt, Nicolas Le Roux, and Francis Bach
        Minimizing Finite Sums with the Stochastic Average Gradient
        https://hal.inria.fr/hal-00860051/document

    SAGA -- Defazio, A., Bach F. & Lacoste-Julien S. (2014).
        SAGA: A Fast Incremental Gradient Method With Support
        for Non-Strongly Convex Composite Objectives
        https://arxiv.org/abs/1407.0202

    Hsiang-Fu Yu, Fang-Lan Huang, Chih-Jen Lin (2011). Dual coordinate descent
        methods for logistic regression and maximum entropy models.
        Machine Learning 85(1-2):41-75.
        http://www.csie.ntu.edu.tw/~cjlin/papers/maxent_dual.pdf
    """

    __slots__ = ('_LogisticRegressionSklearn')

    _param_list = []    # Store original arguments
    _uparams = None     # Store user provided arguments

    # Override the existing function to include the current list of parameters
    # as the init signature is modified by the decorator
    def _get_param_names(self):
        return self._param_list

    @_check_uparams
    def __init__(self, penalty='l2', dual=False, tol=1e-4, C=1.0,
                 fit_intercept=True, intercept_scaling=1, class_weight=None,
                 random_state=None, solver='warn', max_iter=100,
                 multi_class='warn', verbose=0, warm_start=False, n_jobs=None,
                 use_gpu=True,                      # pai4sk
                 device_ids=[],                     # pai4sk
                 return_training_history=None,      # pai4sk
                 privacy=False,                     # pai4sk
                 eta=0.3,                           # pai4sk
                 batch_size=100,                    # pai4sk
                 privacy_epsilon=10,                # pai4sk
                 grad_clip=1,                       # pai4sk
                 num_threads=1):                    # pai4sk

        self.penalty = penalty
        self.dual = dual
        self.tol = tol
        self.C = C
        self.fit_intercept = fit_intercept
        self.intercept_scaling = intercept_scaling
        self.class_weight = class_weight
        self.random_state = random_state
        self.solver = solver
        self.max_iter = max_iter
        self.multi_class = multi_class
        self.verbose = verbose
        self.warm_start = warm_start
        self.n_jobs = n_jobs
        self.use_gpu = use_gpu
        self.device_ids = device_ids
        self.num_threads = num_threads
        self.return_training_history = return_training_history
        self._snaplr = None
        self._LogisticRegressionSklearn = super(LogisticRegression, self)
        self.privacy = privacy
        self.eta = eta
        self.batch_size = batch_size
        self.privacy_epsilon = privacy_epsilon
        self.grad_clip = grad_clip

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._LogisticRegressionSklearn = super(LogisticRegression, self)

    def _check_set_snapml(self, X, y):
        if self.solver is not 'warn' and self.solver is not 'snapml':
            return False

        snap_args = {}
        mod_params = []
        unsupported_params = []

        if ( not self.random_state is None):
            unsupported_params.append("random_state")

        if (type(self.class_weight) is dict):
            unsupported_params.append("class_weight")

        if not _check_snapml_types(X,y):
            if 'solver' in self._uparams and self.solver == 'snapml':
                raise ValueError("SnapML: unsupported data types")
            unsupported_params.append("X/y")
        else:
            if self.solver is not 'snapml' and not _is_data_snapml_partition_type(y) and not _is_gpu_matrix_type(X, y):
                uniq_labs_ = np.unique(y)
                if (len(uniq_labs_) == 2):
                    if not((uniq_labs_[0] == -1 and uniq_labs_[1] == 1) or
                        (uniq_labs_[0] == 1 and uniq_labs_[1] == -1) or
                        (uniq_labs_[0] == 0 and uniq_labs_[1] == 1) or
                        (uniq_labs_[0] == 1 and uniq_labs_[1] == 0)):
                        if 'solver' in self._uparams and self.solver == 'snapml':
                            raise ValueError("SnapML: supported labels for binary classification are {0,1} or {-1,1}.")
                        unsupported_params.append("y")

        # choose CPU implementation for smaller dataset (in-memory size < 400kb)
        if not _is_data_snapml_partition_type(X) and not _is_gpu_matrix_type(X, y):
            if _get_data_size(X) < 400 and not 'use_gpu' in self._uparams :
                self.use_gpu = False
                mod_params.append("use_gpu")

        if not unsupported_params:
            # Check for GPUs
            snap_args['use_gpu'] = False if self.use_gpu is None else self.use_gpu
            snap_args['device_ids'] = [] if self.device_ids is None else self.device_ids
            if self.use_gpu:
                from pai4sk import sml_utils
                ndevices = sml_utils.get_gpu_count()
                if ndevices == 0:
                    if 'use_gpu' in self._uparams:
                        raise ValueError("SnapML: GPU device not found")
                    mod_params.append("use_gpu")
                    self.use_gpu = False

            if 'max_iter' in self._uparams:
                snap_args['max_iter'] = self.max_iter
            else:
                mod_params.append("max_iter")
                snap_args['max_iter'] = 1000

            if 'dual' in self._uparams:
                snap_args['dual'] = self.dual
            else:
                mod_params.append("dual")
                snap_args['dual'] = True

            if 'tol' in self._uparams:
                snap_args['tol'] = self.tol
            else:
                mod_params.append("tol")
                snap_args['tol'] = 0.001

            if 'fit_intercept' in self._uparams:
                snap_args['fit_intercept'] = self.fit_intercept
            else:
                mod_params.append("fit_intercept")
                self.fit_intercept = False

            snap_args['verbose'] = False if self.verbose == 0 else True
            snap_args['regularizer'] = 1 if self.C is None else 1/self.C
            snap_args['intercept_scaling'] = 1.0 if self.intercept_scaling is None else self.intercept_scaling
            snap_args['return_training_history'] = self.return_training_history
            snap_args['penalty'] = self.penalty
            snap_args['privacy'] = self.privacy
            snap_args['eta'] = self.eta
            snap_args['batch_size'] = self.batch_size
            snap_args['privacy_epsilon'] = self.privacy_epsilon
            snap_args['grad_clip'] = self.grad_clip
            snap_args['class_weight'] = self.class_weight
            snap_args['num_threads'] = self.num_threads
        
            from pai4sk import LogisticRegression as _snaplr
            self._snaplr = _snaplr(**snap_args)

            if self.solver == 'warn' and self._snaplr:
                self.solver = 'snapml'
                warnings.warn("SnapML: Default solver is set to 'snapml'. "
                          "Specify a solver to silence this warning.",
                          UserWarning)

        if unsupported_params:
            if 'solver' in self._uparams and self.solver == 'snapml':
                raise ValueError("SnapML: unsupported parameters/value for 'snapml' solver: " + ", ".join([str(x) for x in unsupported_params]))

            warnMsg = "SnapML: fallback to sklearn implementation due to unsupported parameters/values: " + ", ".join([str(x) for x in unsupported_params])
            warnings.warn(warnMsg, UserWarning) 
            self.solver = 'warn'
            return False

        if mod_params:
            warnMsg = "SnapML: Default values for these parameters are modified for 'snapml' solver: " + ", ".join([str(x) for x in mod_params])
            warnings.warn(warnMsg, UserWarning)

        return True

    def fit(self, X, y, sample_weight=None):
        """Fit the model according to the given training data.
        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.
            For SnapML solver it also supports input of types SnapML data partition and DeviceNDArray.
        y : array-like, shape (n_samples,) or (n_samples, n_targets)
            Target vector relative to X.
        sample_weight : array-like, shape (n_samples,) optional
            Array of weights that are assigned to individual samples.
            If not provided, then each sample is given unit weight.
            .. versionadded:: 0.17
               *sample_weight* support to LogisticRegression.
        Returns
        -------
        self : object
        """
        if not isinstance(self.C, numbers.Number) or self.C < 0:
            raise ValueError("Penalty term must be positive; got (C=%r)"
                             % self.C)
        if not isinstance(self.max_iter, numbers.Number) or self.max_iter < 0:
            raise ValueError("Maximum number of iteration must be positive;"
                             " got (max_iter=%r)" % self.max_iter)
        if not isinstance(self.tol, numbers.Number) or self.tol < 0:
            raise ValueError("Tolerance for stopping criteria must be "
                             "positive; got (tol=%r)" % self.tol)

        if self._check_set_snapml(X, y):
            # snap-ml fit
            self.training_history_ = self._snaplr.fit(X, y)

            if not _mpi_enabled:
                # Attributes only supported with snap-ml-local
                self.support_ = self._snaplr.support_
                self.model_sparsity_ = self._snaplr.model_sparsity_
                self.n_iter_ = self._snaplr.n_iter_

            self.classes_ = self._snaplr.uniq_labs_.astype(np.int)
            self.coef_ = self._snaplr.model_
            if (self.fit_intercept):
                self.coef_ = self.coef_[:,:-1]
                self.intercept_ = self.coef_[:,-1]
            else:
                #self.coef_ is already of required shape
                self.intercept_ = np.empty(shape=())

            return self
        else:
            lr_sk = self._LogisticRegressionSklearn.fit(X, y, sample_weight)
            self.classes_ = lr_sk.classes_
            self.coef_ = lr_sk.coef_
            self.intercept_ = lr_sk.intercept_
            self.n_iter_ = lr_sk.n_iter_
            self.training_history_ = []
            return self


    def predict_proba(self, X, num_threads=0):
        """Probability estimates.
        The returned estimates for all classes are ordered by the
        label of classes.
        For a multi_class problem, if multi_class is set to be "multinomial"
        the softmax function is used to find the predicted probability of
        each class.
        Else use a one-vs-rest approach, i.e calculate the probability
        of each class assuming it to be positive using the logistic function.
        and normalize these values across all the classes.
        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            For SnapML solver it also supports input of type SnapML data partition.
        num_threads : int, default : 0
            Number of threads used to run inference.
            By default inference runs with maximum number of available threads.
        Returns
        -------
        T : array-like, shape = [n_samples, n_classes]
            Returns the probability of the sample for each class in the model,
            where classes are ordered as they are in ``self.classes_``.
        """
        if not hasattr(self, "coef_"):
            raise NotFittedError("Call fit before prediction")

        if self.solver == 'snapml' and len(self._snaplr.uniq_labs_) <= 2:
            return self._snaplr.predict_proba(X, num_threads)
        else:
            return self._LogisticRegressionSklearn.predict_proba(X)

    def predict(self, X, num_threads=0):
        """
        Class predictions
        The returned class estimates.
        Parameters
        ----------
        X : sparse matrix (csr_matrix) or dense matrix (ndarray)
            Dataset used for predicting class estimates.
            For SnapML solver it also supports input of type SnapML data partition.
        num_threads : int, default : 0
            Number of threads used to run inference.
            By default inference runs with maximum number of available threads.
        Returns
        -------
        proba: array-like, shape = (n_samples,)
            Returns the predicted class of the sample.
        """
        if not hasattr(self, "coef_"):
            raise NotFittedError("Call fit before prediction")

        if _check_snapml_type(X) and self.solver == 'snapml':
            return self._snaplr.predict(X, num_threads)
        else:
            return self._LogisticRegressionSklearn.predict(X)


    def predict_log_proba(self, X):
        """Log of probability estimates.
        The returned estimates for all classes are ordered by the
        label of classes.
        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            For SnapML solver it also supports input of type SnapML data partition.
        Returns
        -------
        T : array-like, shape = [n_samples, n_classes]
            Returns the log-probability of the sample for each class in the
            model, where classes are ordered as they are in ``self.classes_``.
        """
        if self.solver == 'snapml' and len(self._snaplr.uniq_labs_) <= 2:
            return self._snaplr.predict_log_proba(X)
        else:
            return self._LogisticRegressionSklearn.predict_log_proba(X)





