# Simple Examples

In this example we will train a Linear Regression on the generated dataset, using both `scikit-learn` (`sklearn`) and `pai4sk`.
Note that `scikit-learn` does not provide an MPI implementation. We will however still perform this comparison in order to mainly verify the accuracy metrics.

Please run the following commands to generate data:

```bash
mkdir data/
python gen_data.py
```

## Load pre-processed data

After running the data pre-processing steps above, we are ready to perform our benchmarking: `snap.ml` vs. `sklearn`.
Pre-processed data can be loaded with below methods. See API documentation of the following methods for more details.
```
1.`load_from_snap_format`
See example-simple-snap-format.py
2.`load_snap_file`
See example-simple-snap-file.py
3.`load_from_svmlight_format`
See example-simple-svmlight-format.py
4.`load_svmlight_file`
See example-simple-svmlight-file.py
```

`load_from_snap_format` and `load_from_svmlight_format` return data_ptr.  data_ptr will be of type DensePartition, SparsePartition or ConstantValueSparsePartition.

`load_snap_file` returns dense or sparse matrices as X,y. Here X will be npy, csr or csc and y will be npy.

`load_svmlight_file` returns sparse matrix X and numpy array y. Here X will be csr or csc.

Note: dense snap data can be loaded with `load_from_snap_format` and `load_snap_file` methods.

## Training and Evaluating a Linear Regression Model

First we train a Linear Regression model using the IBM `pai4sk` library. 
We provide command line option --use_gpu to run the example for how to train the model using 1) GPUs or 2) CPUs.

We define the data paths for the train and test datasets in distributed snap format.

You need to have access to the data from all the connected machines that will be used (by replicating the data or by providing a distributed file system).

Next we read the datasets from snap format using the loading functions provided by the `pai4sk` library and we measure the data loading time.

After initialization, we call the `fit()` function in order to train the dual Logistic Regression model. We also time the execution of the training process.

Please note that the `pai4sk` library assumes for maximum performance to run 1 MPI process per node.

As a last step we make predictions on the test data and we compute the accuracy and the logistic loss metrics. 
The `pai4sk` library provides distributed implementations of the accuracy and logistic loss metrics.

Note that the `pai4sk` CPU algorithm converges significantly faster than the
`pai4sk` GPU algorithm with respect to the number of epochs (iterations). Therefore
it is sufficient to set the `max_iter` parameter to 6.
Moreover, the CPU algorithm supports multi-threading training. To set the number of threads change the `num_threads` variable. 

If the CPU training does not converge, set the `num_threads` to 1 or use the GPU algorithm.

Then we train a Logistic Regression model using the `sklearn` library.

## Results

To run on one machine with 1 MPI process, one should run the following command: 
```bash
snaprun python $CONDA_PREFIX/pai4sk/mpi-examples/example-simple/example-simple-snap-file.py
```

To run on 2 machines with 1 MPI process per machine, one should run the following command:
```bash
snaprun -H host1,host2 python $CONDA_PREFIX/pai4sk/mpi-examples/example-simple/example-simple-snap-file.py
```

The user can also run multiple MPI processes per machine, but, in that case, the MPI processes on one machine will share the same hardware resources during training. Namely, if the user trains the model on GPUs, the MPI processes on one machine will share the same GPUs as defined in the `device_ids` parameter. Therefore, we recommend the user to run 1 MPI process per machine in order to achieve the maximum performance.


One should see the following results on 1 **P9+V100 machine with 4 GPUs** when running `snaprun python $CONDA_PREFIX/pai4sk/mpi-examples/example-simple/example-simple-snap-file.py`:
```
[pai4sk] Summary:
[pai4sk] [Data loading] Time = 3.5649
[pai4sk] [GPU Training] Accuracy = 0.7877
[pai4sk] [GPU Training] LogLoss = 0.4558
[pai4sk] [GPU Training] Training time = 56.7975

[Sklearn] Summary:
[Sklearn] [Data loading] Time = 36.6448
[Sklearn] Accuracy = 0.7877
[Sklearn] LogLoss = 0.4558
[Sklearn] Training time = 1564.0317
```

One should see the following results on 2 **P9+V100 machines each with 4 GPUs** connected with Infiniband, when running `snaprun -H host1,host2 python $CONDA_PREFIX/pai4sk/mpi-examples/example-simple/example-simple-snap-file.py`:
```
[pai4sk] Summary:
[pai4sk] [Data loading] Time = 1.9235
[pai4sk] [GPU Training] Accuracy = 0.7877
[pai4sk] [GPU Training] LogLoss = 0.4558
[pai4sk] [GPU Training] Training time = 37.1040

[Sklearn] Summary:
[Sklearn] [Data loading] Time = 36.6448
[Sklearn] Accuracy = 0.7877
[Sklearn] LogLoss = 0.4558
[Sklearn] Training time = 1564.0317
```

Please note that the reported times can vary depending on the system architecture.

&copy; Copyright IBM Corporation 2019
